
view_set_default = (0, 1, 2)
num_slices_default = (10,)
sampler_default = 'linear'

min_density_default = 0.01

num_rows_per_view_default = 2

figsize_default=(14, 10)
bounding_rect_default=(0.02, 0.02, 0.98, 0.98)

num_cols_grid_default = 6
grid_pad_default = 0.01
axis_pad_default = 0.01

background_value = 0
